from ticlib import TicI2C
import time
from MM_range_calculator import range_calculator
import MM_byte_parser

#funkce pro kontrolu driveru, jestli je ve stavu energized
def driver_is_energized(driver: TicI2C):
    driver_misc = 0
    for x in range(10):
        try:
            driver_misc = MM_byte_parser.misc_flag_format(driver.get_misc_flags())
            break
        except:
            pass

    if driver_misc != 0:
        return bool(driver_misc["Energized"])
    else:
        return False

#třída osy, která implementuje jednotlivé drivery a funkce
class Axis(object):
    #inicializace 
    def __init__(self, name: str, driver1: TicI2C, driver2: TicI2C):
        self.name = name
        self.driver1 = driver1
        self.driver2 = driver2
        self.initialized = 0

        self.motors = {
        "D1Control": 0,
        "D2Control": 0}

        self.state_var = {
        "TargetPosition": 0,
        "TargetPositionMM": 0,
        "ActualPosition": 0,
        "ActualPositionMM": 0,
        "MaxSpeed": 100000000,
        "StartingSpeed": 1000000,
        "MaxAcceleration": 500000,
        "MaxDeceleration": 500000,
        "Control": 0,
        "D1Active": 0,
        "D2Active": 0,
        "AxisRangeAbs": 0,
        "RangeMM": 0,}

    #metoda pro čištění errorů driverů a nastavení do provozního stavu
    def clear_axis_error(self):
        try:
            self.driver1.clear_driver_error()
            self.driver1.energize()
            self.driver1.exit_safe_start()
        except:
            pass
        try:
            self.driver2.clear_driver_error()
            self.driver2.energize()
            self.driver2.exit_safe_start()
        except:
            pass

    #metoda pro kontrolu aktivních motorů
    def is_active(self):
        motor_1 = int(self.motors["D1Control"])
        motor_2 = int(self.motors["D2Control"])
        if motor_1 == 1 or motor_2 == 1:
            return True
        else:
            return False

    #metoda pro nastavení polohy osy v absolutní vzdálenosti -100 až 100 procent z rozsahu
    def set_target_position(self, target_position: float):
        if -100.00 <= float(target_position) <= 100.00: #kontrola, zda poloha není mimo rozsah
            target_position = round(target_position, 2)  #float na dvě desetinná místa
            range = self.state_var["AxisRangeAbs"] #načtení rozsahu z parametru instance (osy)
            tp = (range / 100) * target_position #přepočet polohy na absolutní pozici v krocích motoru
            tp = int(tp) #ořezání desetinných čísel převedením na int
            try:
                if self.can_i_drive() and self.has_two_active_motors(): #pokud je driver aktivní (oba), poslat příkaz na nastavení cílové polohy
                    self.driver1.set_target_position(tp) 
                    self.driver2.set_target_position(tp)
                elif self.can_i_drive() and (not self.has_two_active_motors()): #pokud je driver aktivní (jeden), poslat příkaz na nastavení cílové polohy
                    self.driver1.set_target_position(tp)
                else:
                    pass
            except:
                pass
    
    #metoda pro dotaz, zda má osa jeden nebo dva motory
    def has_two_active_motors(self):
        motor_1 = int(self.motors["D1Control"])
        motor_2 = int(self.motors["D2Control"])
        if motor_1 == 1 & motor_2 == 1:
            return True
        else:
            return False

    #metoda pro dotaz, aktuální pozice motoru (v úvahu pouze hlavní motor 1, synchronní osa v režimu STEP/DIR nemá informaci o poloze)
    def get_current_position(self):
        out = 0
        for x in range(10):
            try:
                out = self.driver1.get_current_position()
                out = ((out * 100) / self.state_var["AxisRangeAbs"]) #převedení hodnoty v krocích na procenta od -100 do 100
                out = float(round(out,2)) #zaokrouhlení na dvě desetinná místa
                break
            except:
                pass
        return out

    #metoda pro dotaz, zda jsou připojené motory schnopny provozu
    def can_i_drive(self) -> bool:
        active_driver_1 = self.is_driver_active(self.driver1)
        active_driver_2 = self.is_driver_active(self.driver2)
        if not (active_driver_1 or active_driver_2):
            return False

        if active_driver_1 and (not active_driver_2):
            driver1_is_energized = driver_is_energized(self.driver1)
            driver1_has_errors = self.driver_has_errors(self.driver1)

            if driver1_is_energized and (not driver1_has_errors):
                return True

        elif active_driver_1 & active_driver_2:
            driver1_is_energized = driver_is_energized(self.driver1)
            driver1_has_errors = self.driver_has_errors(self.driver1)

            driver2_is_energized = driver_is_energized(self.driver2)
            driver2_has_errors = self.driver_has_errors(self.driver2)

            if driver1_is_energized and (not driver1_has_errors) and driver2_is_energized and (not driver2_has_errors):
                return True
            else:
                return False
        else:
            return False

    @staticmethod #metoda pro dotaz, zda má driver nějaký error
    def driver_has_errors(driver: TicI2C) -> bool:
        driver_errors = 0
        for x in range(10):
            try:
                driver_errors = MM_byte_parser.error_status_format(driver.get_error_status())
                break
            except:
                pass
        without_errors = all(value == 0 for value in driver_errors.values())
        return not without_errors

    @staticmethod #metoda pro dotaz, zda driver komnikuje -> je aktivní
    def is_driver_active(driver: TicI2C) -> bool:
        is_ok = False
        count = 0
        driver_misc_flag = 0
        while not is_ok:
            if count == 10:
                break
            try:
                driver_misc_flag = driver.get_misc_flags() #pokud driver odpoví, je aktivní, nepoužívají se příchozí data
                is_ok = True
            except:
                driver_misc_flag = 0
                count += 1

        if driver_misc_flag != 0:
            return True
        else:
            return False

    #metoda, která driver nastaví do provozuschopného stavu 
    def setup_driver_for_ride(self) -> bool:
        for x in range(10):
            try:
                # nastaveni vychozich hodnot a spusteni motoru
                self.driver1.clear_driver_error()
                self.driver1.energize()
                self.driver1.exit_safe_start()
                self.driver1.set_max_acceleration(self.state_var["MaxAcceleration"])
                self.driver1.set_max_deceleration(self.state_var["MaxDeceleration"])
                self.driver1.set_max_speed(self.state_var["MaxSpeed"])
                self.driver1.set_starting_speed(self.state_var["StartingSpeed"])
                self.driver1.set_step_mode(3) #nastevní kroku na 1/8 kroku (kód 3, dle manuálu)
                return True
            except:
                pass
        return False

    #metoda pro nastavení aktuální pozice motoru jako 0
    def halt_and_set_position(self,set) -> bool:
        for x in range(10):
            try:
                self.driver1.halt_and_set_position(set) 
                return True
            except:
                pass
        return False

    #metoda pro spuštění fce go home, 0 od motoru, 1 k motoru
    def driver_go_home(self,to) -> bool:
        for x in range(10):
            try:
                self.driver1.go_home(to)  
                return True
            except:
                pass
        return False

    #metoda pro testování aktuální pozice při zapnuté fci homeing
    def test_way_go_home(self,target,go_home_set):
        test = 0
        count = 0
        pole = list()
        while True:
            try:
                pos = self.driver1.get_current_position()  #testování aktuální pozice, když dojede na konec = 0 a můžu pokračovat
                pole.append(pos)
                if pos == target: 
                    return pole

                if test != pos: #kontrola, zda se motor po cestě nezasekl, pokud ano, znovu pošlu příkaz co má udělat
                    test = pos
                else:
                    count += 1
                    self.setup_driver_for_ride(self)
                    self.driver1.go_home(go_home_set)
                if count > 100:
                    return False
                time.sleep(0.1)
            except:
                pass
    
    #metoda pro testování aktuální pozice při zapnuté fci homeing
    def test_way_go_position(self,target):
        test = 0
        count = 0
        pole = list()
        while True:
            try:
                pos = self.driver1.get_current_position() #testování aktuální pozice
                pole.append(pos)
                if pos == target: 
                    return pole

                if test != pos: #kontrola, zda se motor po cestě nezasekl, pokud ano, znovu pošlu příkaz co má udělat
                    test = pos
                else:
                    count += 1
                    self.setup_driver_for_ride(self)
                    self.driver1.set_target_position(target)
                if count > 100:
                    return False
                time.sleep(0.1)
            except:
                pass

    #přímé nastavení cílové pozice bez uživatelského zadání
    def driver1_set_target_position(self,target) -> bool:
        for x in range(10):
            try:
                self.driver1.set_target_position(target)
                return True
            except Exception as e:
                pass
        return False

    #metoda pro inicializaci
    def initialize_position(self) -> bool:
        if self.setup_driver_for_ride() == False: return False #zastav, pokud motor není schopen jízdy
        if self.halt_and_set_position(0) == False: return False #zastav, pokud se nepodaří vynulovat pozici

        go_home = 1 #go home 1, k motoru
        if self.driver_go_home(go_home) == False: return False #zastav, pokud se nepodaří dojet na konec jedné ze stran "domů"
        
        time.sleep(2)  # čekání na pohnutí motoru z výchozí pozice
        if self.test_way_go_home(0,go_home) == False: return False #testování, zda motor dokončil cestu

        go_home = 0 #go home 0, na konec
        if self.driver_go_home(go_home) == False: return False #zastav, pokud se nepodaří dojet na konec jedné ze stran "domů"

        time.sleep(2)  # čekání na pohnutí motoru z výchozí pozice
        pole = self.test_way_go_home(0,go_home) 
        if pole == False: return False #testování, zda motor dokončil cestu, pokud ano, mám pole s daty o průběžné poloze, pro výpočet rozsahu osy

        range_value = range_calculator(pole) #výpočet rozsahu osy ze vstupních dat
        if range_value == 0: return False #vypni inicializaci, pokud se něco nepovedlo

        center = int(range_value / 2)
        self.state_var["AxisRangeAbs"] = range_value / 2
        if self.driver1_set_target_position(center) == False: return False #nastavení pozice motoru do středu osy

        time.sleep(2)
        if self.test_way_go_position(center) == False: return False #testování, zda motor dokončil cestu
        if self.halt_and_set_position(0) == False: return False #neúspěžné, pokud se nepodaří vynulovat pozici

        center = 0
        return True

    #metoda pro debugging, neslouží pro běžný běh programu
    def kontrola(self):

        # General status  -------------------------------------
        print("\nGeneral status")
        print("get_error_occured: ", self.driver1.get_error_occured())
        print("get_error_status: ", MM_byte_parser.error_status_format(self.driver1.get_error_status()))
        print("get_misc_flags: ", MM_byte_parser.misc_flag_format(self.driver1.get_misc_flags()))
        print("get_operation_state: ", self.driver1.get_operation_state())

        # Step planning ---------------------------------------
        print("\nStep planning")
        print("get_acting_target_position: ", self.driver1.get_acting_target_position())
        print("get_current_position: ", self.driver1.get_current_position())
        print("get_current_velocity: ", self.driver1.get_current_velocity())
        print("get_max_acceleration: ", self.driver1.get_max_acceleration())
        print("get_max_deceleration: ", self.driver1.get_max_deceleration())
        print("get_max_speed: ", self.driver1.get_max_speed())
        print("get_planning_mode: ", self.driver1.get_planning_mode())
        print("get_starting_speed: ", self.driver1.get_starting_speed())
        print("get_target_position: ", self.driver1.get_target_position())
        print("get_target_velocity: ", self.driver1.get_target_velocity())
        print("get_time_since_last_step: ", self.driver1.get_time_since_last_step())

        # Other -----------------------------------------------
        print("\nOther")
        print("get_current_limit: ", self.driver1.get_current_limit())
        print("get_device_reset: ", self.driver1.get_device_reset())
        print("get_step_mode: ", self.driver1.get_step_mode())
        print("get_vin_voltage [V]: ", self.driver1.get_vin_voltage() / 1000)
        print("get_uptime [min]: ", (self.driver1.get_uptime() / 1000) / 60)
        print("get_control_mode: ", self.driver1.settings.get_control_mode())

        # Miscellaneous -------------------------------------------------
        print("\nMiscellaneous")
        print("get_auto_clear_driver_error: ", self.driver1.settings.get_auto_clear_driver_error())
        print("get_disable_safe_start: ", self.driver1.settings.get_disable_safe_start())
        print("get_ignore_err_line_high: ", self.driver1.settings.get_ignore_err_line_high())
        print("get_never_sleep: ", self.driver1.settings.get_never_sleep())
        print("get_vin_calibration: ", self.driver1.settings.get_vin_calibration())

        # Soft error response -------------------------------------------
        print("\nSoft error response")
        print("get_current_limit_during_error: ", self.driver1.settings.get_current_limit_during_error())
        print("get_soft_error_position: ", self.driver1.settings.get_soft_error_position())
        print("get_soft_error_response: ", self.driver1.settings.get_soft_error_response())

        # Serial --------------------------------------------------------
        print("\nSerial")
        print("get_serial_7bit_responses: ", self.driver1.settings.get_serial_7bit_responses())
        print("get_serial_14bit_device_number: ", self.driver1.settings.get_serial_14bit_device_number())
        print("get_serial_alt_device_number: ", self.driver1.settings.get_serial_alt_device_number())
        print("get_serial_baud_rate: ", self.driver1.settings.get_serial_baud_rate())
        print("get_serial_command_timeout: ", self.driver1.settings.get_serial_command_timeout())
        print("get_serial_crc_for_commands: ", self.driver1.settings.get_serial_crc_for_commands())
        print("get_serial_crc_for_responses: ", self.driver1.settings.get_serial_crc_for_responses())
        print("get_serial_device_number: ", self.driver1.settings.get_serial_device_number())
        print("get_serial_enable_alt_device_number: ", self.driver1.settings.get_serial_enable_alt_device_number())
        print("get_serial_response_delay: ", self.driver1.settings.get_serial_response_delay())

        # Pin Configuration ---------------------------------------------
        print("\nPin Configuration")
        # SCL
        print("\nSCL")
        print("get_scl_active_high: ", self.driver1.settings.get_scl_active_high())
        print("get_scl_config: ", self.driver1.settings.get_scl_config())
        print("get_scl_enable_analog: ", self.driver1.settings.get_scl_enable_analog())
        print("get_scl_enable_pull_up: ", self.driver1.settings.get_scl_enable_pull_up())
        print("get_scl_kill_switch: ", self.driver1.settings.get_scl_kill_switch())
        print("get_scl_limit_switch_forward: ", self.driver1.settings.get_scl_limit_switch_forward())
        print("get_scl_limit_switch_reverse: ", self.driver1.settings.get_scl_limit_switch_reverse())
        print("get_scl_pin_function: ", self.driver1.settings.get_scl_pin_function())
        # SDA
        print("\nSDA")
        print("get_sda_active_high: ", self.driver1.settings.get_sda_active_high())
        print("get_sda_config: ", self.driver1.settings.get_sda_config())
        print("get_sda_enable_analog: ", self.driver1.settings.get_sda_enable_analog())
        print("get_sda_enable_pull_up: ", self.driver1.settings.get_sda_enable_pull_up())
        print("get_sda_kill_switch: ", self.driver1.settings.get_sda_kill_switch())
        print("get_sda_limit_switch_forward: ", self.driver1.settings.get_sda_limit_switch_forward())
        print("get_sda_limit_switch_reverse: ", self.driver1.settings.get_sda_limit_switch_reverse())
        print("get_sda_pin_function: ", self.driver1.settings.get_sda_pin_function())
        # TX
        print("\nTX")
        print("get_tx_active_high: ", self.driver1.settings.get_tx_active_high())
        print("get_tx_config: ", self.driver1.settings.get_tx_config())
        print("get_tx_enable_analog: ", self.driver1.settings.get_tx_enable_analog())
        print("get_tx_kill_switch: ", self.driver1.settings.get_tx_kill_switch())
        print("get_tx_limit_switch_forward: ", self.driver1.settings.get_tx_limit_switch_forward())
        print("get_tx_limit_switch_reverse: ", self.driver1.settings.get_tx_limit_switch_reverse())
        print("get_tx_pin_function: ", self.driver1.settings.get_tx_pin_function())
        # RX
        print("\nRX")
        print("get_rx_active_high: ", self.driver1.settings.get_rx_active_high())
        print("get_rx_config: ", self.driver1.settings.get_rx_config())
        print("get_rx_enable_analog: ", self.driver1.settings.get_rx_enable_analog())
        print("get_rx_kill_switch: ", self.driver1.settings.get_rx_kill_switch())
        print("get_rx_limit_switch_forward: ", self.driver1.settings.get_rx_limit_switch_forward())
        print("get_rx_limit_switch_reverse: ", self.driver1.settings.get_rx_limit_switch_reverse())
        print("get_rx_pin_function: ", self.driver1.settings.get_rx_pin_function())
        # RC
        print("\nRC")
        print("get_rc_active_high: ", self.driver1.settings.get_rc_active_high())
        print("get_rc_config: ", self.driver1.settings.get_rc_config())
        print("get_rc_kill_switch: ", self.driver1.settings.get_rc_kill_switch())
        print("get_rc_limit_switch_forward: ", self.driver1.settings.get_rc_limit_switch_forward())
        print("get_rc_limit_switch_reverse: ", self.driver1.settings.get_rc_limit_switch_reverse())

        # Motor ---------------------------------------------------------
        print("\nMotor")
        print("get_current_limit: ", self.driver1.settings.get_current_limit())
        print("get_decay_mode: ", self.driver1.settings.get_decay_mode())
        print("get_invert_motor_direction: ", self.driver1.settings.get_invert_motor_direction())
        print("get_max_acceleration: ", self.driver1.settings.get_max_acceleration())
        print("get_max_deceleration: ", self.driver1.settings.get_max_deceleration())
        print("get_max_speed: ", self.driver1.settings.get_max_speed())
        print("get_starting_speed: ", self.driver1.settings.get_starting_speed())
        print("get_step_mode: ", self.driver1.settings.get_step_mode())

        # Homing --------------------------------------------------------
        print("\nHoming")
        print("get_auto_homing: ", self.driver1.settings.get_auto_homing())
        print("get_auto_homing_forward: ", self.driver1.settings.get_auto_homing_forward())
        print("get_homing_speed_away: ", self.driver1.settings.get_homing_speed_away())
        print("get_homing_speed_towards: ", self.driver1.settings.get_homing_speed_towards())


#seznam možných příkazů pro pololu drivery
"""
Available commands:#############################################################

tic.clear_driver_error()
tic.deenergize()
tic.energize()
tic.enter_safe_start()
tic.exit_safe_start()
tic.go_home(value)
tic.halt_and_hold()
tic.halt_and_set_position(value)
tic.reset()
tic.reset_command_timeout()
tic.set_agc_option(value)
tic.set_current_limit(value)
tic.set_decay_mode(value)
tic.set_max_acceleration(value)
tic.set_max_deceleration(value)
tic.set_max_speed(value)
tic.set_starting_speed(value)
tic.set_step_mode(value)
tic.set_target_position(value)
tic.set_target_velocity(value)

Available variables:#############################################################

# General status  -------------------------------------
tic.get_error_occured()
tic.get_error_status()
tic.get_misc_flags()
tic.get_operation_state()

# Step planning ---------------------------------------
tic.get_acting_target_position()
tic.get_current_position()
tic.get_current_velocity()
tic.get_max_acceleration()
tic.get_max_deceleration()
tic.get_max_speed()
tic.get_planning_mode()
tic.get_starting_speed()
tic.get_target_position()
tic.get_target_velocity()
tic.get_time_since_last_step()

# Other -----------------------------------------------
tic.get_analog_reading_rx()
tic.get_analog_reading_scl()
tic.get_analog_reading_sda()
tic.get_analog_reading_tx()
tic.get_current_limit()
tic.get_decay_mode()
tic.get_device_reset()
tic.get_digital_readings()
tic.get_encoder_position()
tic.get_input_after_averaging()
tic.get_input_after_hysteresis()
tic.get_input_after_scaling()
tic.get_input_state()
tic.get_pin_states()
tic.get_rc_pulse()
tic.get_step_mode()
tic.get_vin_voltage()
tic.get_uptime()

# T249-only -------------------------------------------
tic.get_agc_bottom_current_limit()
tic.get_agc_current_boost_steps()
tic.get_agc_frequency_limit()
tic.get_agc_mode()
tic.get_last_motor_driver_error()

# 36v4-only -------------------------------------------
tic.get_last_hp_driver_errors()

Available settings:#############################################################

tic.settings.get_control_mode()

# Miscellaneous -------------------------------------------------
tic.settings.get_auto_clear_driver_error()
tic.settings.get_disable_safe_start()
tic.settings.get_ignore_err_line_high()
tic.settings.get_never_sleep()
tic.settings.get_vin_calibration()

# Soft error response -------------------------------------------
tic.settings.get_current_limit_during_error()
tic.settings.get_soft_error_position()
tic.settings.get_soft_error_response()

# Serial --------------------------------------------------------
tic.settings.get_serial_7bit_responses()
tic.settings.get_serial_14bit_device_number()
tic.settings.get_serial_alt_device_number()
tic.settings.get_serial_baud_rate()
tic.settings.get_serial_command_timeout()
tic.settings.get_serial_crc_for_commands()
tic.settings.get_serial_crc_for_responses()
tic.settings.get_serial_device_number()
tic.settings.get_serial_enable_alt_device_number()
tic.settings.get_serial_response_delay()

# Encoder -------------------------------------------------------
tic.settings.get_encoder_postscaler()
tic.settings.get_encoder_prescaler()
tic.settings.get_encoder_unlimited()

# Input conditioning --------------------------------------------
tic.settings.get_input_averaging_enabled()
tic.settings.get_input_hysteresis()

# RC and analog scaling -----------------------------------------
tic.settings.get_input_invert()
tic.settings.get_input_max()
tic.settings.get_input_min()
tic.settings.get_input_neutral_max()
tic.settings.get_input_neutral_min()
tic.settings.get_input_scaling_degree()
tic.settings.get_output_max()
tic.settings.get_output_min()

# Pin Configuration ---------------------------------------------
# SCL
tic.settings.get_scl_active_high()
tic.settings.get_scl_config()
tic.settings.get_scl_enable_analog()
tic.settings.get_scl_enable_pull_up()
tic.settings.get_scl_kill_switch()
tic.settings.get_scl_limit_switch_forward()
tic.settings.get_scl_limit_switch_reverse()
tic.settings.get_scl_pin_function()
# SDA
tic.settings.get_sda_active_high()
tic.settings.get_sda_config()
tic.settings.get_sda_enable_analog()
tic.settings.get_sda_enable_pull_up()
tic.settings.get_sda_kill_switch()
tic.settings.get_sda_limit_switch_forward()
tic.settings.get_sda_limit_switch_reverse()
tic.settings.get_sda_pin_function()
# TX
tic.settings.get_tx_active_high()
tic.settings.get_tx_config()
tic.settings.get_tx_enable_analog()
tic.settings.get_tx_kill_switch()
tic.settings.get_tx_limit_switch_forward()
tic.settings.get_tx_limit_switch_reverse()
tic.settings.get_tx_pin_function()
# RX
tic.settings.get_rx_active_high()
tic.settings.get_rx_config()
tic.settings.get_rx_enable_analog()
tic.settings.get_rx_kill_switch()
tic.settings.get_rx_limit_switch_forward()
tic.settings.get_rx_limit_switch_reverse()
tic.settings.get_rx_pin_function()
# RC
tic.settings.get_rc_active_high()
tic.settings.get_rc_config()
tic.settings.get_rc_kill_switch()
tic.settings.get_rc_limit_switch_forward()
tic.settings.get_rc_limit_switch_reverse()

# Motor ---------------------------------------------------------
tic.settings.get_current_limit()
tic.settings.get_decay_mode()
tic.settings.get_invert_motor_direction()
tic.settings.get_max_acceleration()
tic.settings.get_max_deceleration()
tic.settings.get_max_speed()
tic.settings.get_starting_speed()
tic.settings.get_step_mode()

# Homing --------------------------------------------------------
tic.settings.get_auto_homing()
tic.settings.get_auto_homing_forward()
tic.settings.get_homing_speed_away()
tic.settings.get_homing_speed_towards()

# T249-only -----------------------------------------------------
tic.settings.get_agc_bottom_current_limit()
tic.settings.get_agc_current_boost_steps()
tic.settings.get_agc_mode()
tic.settings.get_agc_frequency_limit()

# 36v4-only -----------------------------------------------------
tic.settings.get_hp_current_trip_blanking_time()
tic.settings.get_hp_decay_mode()
tic.settings.get_hp_enable_adaptive_blanking_time()
tic.settings.get_hp_enable_unrestricted_current_limits()
tic.settings.get_hp_fixed_off_time()
tic.settings.get_hp_mixed_decay_transition_time()

"""
