import paho.mqtt.client as mqtt
from MM_slave_counter import slave_counter_i2c
from MM_Axis_class import Axis
from MM_switcher_axis import switcher_axis
from MM_switcher_system import switcher_system, slave_driver_control
from MM_initialization import system_info, axis_X, axis_Y, axis_Z, mqtt_init_public,mqtt_init_public_control
from MM_mqtt import client, topic_path_main, topic_path_control, topic_path_classification
import time
import threading

port = 1883
host = "localhost"

#Status inicializace systému
initialized: bool = False

#rozparsování topicu příchozí zprávy 
def function(topic, value):
    topic = topic.split("/")
    if len(topic) != 5: return 0
    try:
        d = {
            "main": topic[0],
            "type": topic[1],
            "classification": topic[2],
            "name": topic[3],
            "variable": topic[4]}
    except:
        pass
    
    #spuštění příslušného switcheru dle typu tématu
    if d.get("type") == "System":
        switcher_system(d.get("variable"), d.get("name"), value)
    elif d.get("type") == "Control":
        switcher_axis(d.get("variable"), d.get("name"), value)

#callback funkce, která se zavolá při připojení MQTT serveru
def on_connect(client, userdata, flags, rc):
    client.subscribe(f'{topic_path_main}/#', 0)

#callback funkce, která se zavolá při přijmutí zprávy
def on_message(client, userdata, msg):
    if system_info.can_i_run == True:
        function(msg.topic, msg.payload.decode("utf_8")) #spuštění funkce pro rozparsování a výběru, kam se mají data poslat

#vypnutí smyčky, když se odpojí MQTT server
def on_disconnect(client, userdata, rc=0):
    client.loop_stop()

#propojení callback fcí
client.on_connect = on_connect
client.on_message = on_message

#přípojení k MQTT serveru
client.connect(host, 1883)

#sledování aktuální polohy ve vlastníím vlákně, nezávisle (asynchronně) na běhu hlavního programu
def get_actual_position(client: mqtt.Client, axis: Axis):
    axis.clear_axis_error() #průběžné mazání errorů, které zastavují motory (zejména serial error)
    current_position = axis.get_current_position() #zjištění aktuální pozice
    current_position_mm = round((int(axis.state_var["RangeMM"])/2) * (current_position/100),2) #přepočet aktuální pozice na vzdálenost v mm
    
    var_topic_ap = f'{topic_path_main}/{topic_path_control}/{topic_path_classification}/{axis.name}/ActualPosition'
    var_topic_ap_mm = f'{topic_path_main}/{topic_path_control}/{topic_path_classification}/{axis.name}/ActualPositionMM'
    client.publish(var_topic_ap, str(current_position)) #odeslání aktuální pozice do MQTT
    client.publish(var_topic_ap_mm, str(current_position_mm)) #odeslání aktuální pozice do MQTT
    # Assign Actual Position
    axis.state_var["ActualPosition"] = float(current_position) #záznam aktuální pozice do parametru dané osy
    axis.state_var["ActualPositionMM"] = float(current_position_mm) #záznam aktuální pozice do parametru dané osy
    threading.Timer(0.5, get_actual_position, (client, axis)).start() #vytvoření nového vlákna, která spustí tuto metodu každých 500ms

#sledování připojených zařízení na i2c, nezávisle (asynchronně) na běhu hlavního programu
def slave_info(client):
    device = slave_counter_i2c(0)
    msg = str(device).replace("{","").replace("}","").replace("'","").replace(",",";")
    client.publish("MIXANM/System/I2C/Info/SlaveI2CCounter",str(msg))
    slave_driver_control()
    threading.Timer(1, slave_info, (client,)).start() #vytvoření nového vlákna, která spustí tuto metodu každou 1s

#inicializační proces při prvním spuštění programu
if __name__ == "__main__":
    time.sleep(2)
    mqtt_init_public(client) #inicializace dat MQTT
    mqtt_init_public_control(client) #inicializace dat MQTT
    #odeslání stavu systému do MQTT
    msg = "Please, initialize the system"
    client.publish("MIXANM/System/I2C/Info/StatusBar", str(msg))
    client.publish("MIXANM/System/I2C/Info/Inicialization", str(0))
    client.publish("MIXANM/System/I2C/Info/InicializedAddress","-")
    client.loop_start() #start mqtt asynchronní smyčky pro přijímání zpráv
    time.sleep(1) #pauza pro očištění příchozích nechtěných dat po připojení
    system_info.can_i_run = True #puštění příchozích zpráv dále do systému
    
    #spuštění průběžné zjišťování polohy pro všechny osy
    for axis in [axis_X, axis_Y, axis_Z]:
        get_actual_position(client, axis)
    slave_info(client)